package javax.xml.validation;

import java.net.URL;
import java.util.ServiceConfigurationError;
import java.util.Iterator;
import java.util.ServiceLoader;
import java.security.AccessController;
import java.security.PrivilegedAction;
import java.security.AccessControlContext;
import com.sun.org.apache.xerces.internal.jaxp.validation.XMLSchemaFactory;
import java.io.File;
import java.util.Properties;

class SchemaFactoryFinder
{
    private static boolean debug;
    private static final SecuritySupport ss;
    private static final String DEFAULT_PACKAGE = "com.sun.org.apache.xerces.internal";
    private static final Properties cacheProps;
    private static volatile boolean firstTime;
    private final ClassLoader classLoader;
    private static final Class<SchemaFactory> SERVICE_CLASS;
    
    private static void debugPrintln(String msg) {
        if (SchemaFactoryFinder.debug) {
            System.err.println("JAXP: " + msg);
        }
    }
    
    public SchemaFactoryFinder(ClassLoader loader) {
        this.classLoader = loader;
        if (SchemaFactoryFinder.debug) {
            this.debugDisplayClassLoader();
        }
    }
    
    private void debugDisplayClassLoader() {
        try {
            if (this.classLoader == SchemaFactoryFinder.ss.getContextClassLoader()) {
                debugPrintln("using thread context class loader (" + this.classLoader + ") for search");
                return;
            }
        }
        catch (Throwable t) {}
        if (this.classLoader == ClassLoader.getSystemClassLoader()) {
            debugPrintln("using system class loader (" + this.classLoader + ") for search");
            return;
        }
        debugPrintln("using class loader (" + this.classLoader + ") for search");
        return;
    }
    
    public SchemaFactory newFactory(String schemaLanguage) {
        if (schemaLanguage == null) {
            throw new NullPointerException();
        }
        SchemaFactory f = this._newFactory(schemaLanguage);
        if (f != null) {
            debugPrintln("factory '" + f.getClass().getName() + "' was found for " + schemaLanguage);
            return f;
        }
        debugPrintln("unable to find a factory for " + schemaLanguage);
        return f;
    }
    
    private SchemaFactory _newFactory(String schemaLanguage) {
        String propertyName = SchemaFactoryFinder.SERVICE_CLASS.getName() + ":" + schemaLanguage;
        try {
            debugPrintln("Looking up system property '" + propertyName + "'");
            String r = SchemaFactoryFinder.ss.getSystemProperty(propertyName);
            if (r != null) {
                debugPrintln("The value is '" + r + "'");
                SchemaFactory sf = this.createInstance(r);
                if (sf != null) {
                    return sf;
                }
            }
            else {
                debugPrintln("The property is undefined.");
            }
        }
        catch (Throwable t) {
            if (SchemaFactoryFinder.debug) {
                debugPrintln("failed to look up system property '" + propertyName + "'");
                t.printStackTrace();
            }
        }
        String javah = SchemaFactoryFinder.ss.getSystemProperty("java.home");
        String configFile = javah + File.separator + "lib" + File.separator + "jaxp.properties";
        try {
            if (SchemaFactoryFinder.firstTime) {
                synchronized (SchemaFactoryFinder.cacheProps) {
                    if (SchemaFactoryFinder.firstTime) {
                        File f = new File(configFile);
                        SchemaFactoryFinder.firstTime = false;
                        if (SchemaFactoryFinder.ss.doesFileExist(f)) {
                            debugPrintln("Read properties file " + f);
                            SchemaFactoryFinder.cacheProps.load(SchemaFactoryFinder.ss.getFileInputStream(f));
                        }
                    }
                }
            }
            String factoryClassName = SchemaFactoryFinder.cacheProps.getProperty(propertyName);
            debugPrintln("found " + factoryClassName + " in $java.home/jaxp.properties");
            if (factoryClassName != null) {
                SchemaFactory sf = this.createInstance(factoryClassName);
                if (sf != null) {
                    return sf;
                }
            }
        }
        catch (Exception ex) {
            if (SchemaFactoryFinder.debug) {
                ex.printStackTrace();
            }
        }
        SchemaFactory factoryImpl = this.findServiceProvider(schemaLanguage);
        if (factoryImpl != null) {
            return factoryImpl;
        }
        if (schemaLanguage.equals("http://www.w3.org/2001/XMLSchema")) {
            debugPrintln("attempting to use the platform default XML Schema validator");
            return new XMLSchemaFactory();
        }
        debugPrintln("all things were tried, but none was found. bailing out.");
        return null;
    }
    
    private Class<?> createClass(String className) {
        boolean internal = false;
        if (System.getSecurityManager() != null && className != null && className.startsWith("com.sun.org.apache.xerces.internal")) {
            internal = true;
        }
        Class<?> clazz;
        try {
            if (this.classLoader != null && !internal) {
                clazz = Class.forName(className, false, this.classLoader);
            }
            else {
                clazz = Class.forName(className);
            }
        }
        catch (Throwable t) {
            if (SchemaFactoryFinder.debug) {
                t.printStackTrace();
            }
            return null;
        }
        return clazz;
    }
    
    SchemaFactory createInstance(String className) {
        SchemaFactory schemaFactory = null;
        debugPrintln("createInstance(" + className + ")");
        Class<?> clazz = this.createClass(className);
        if (clazz == null) {
            debugPrintln("failed to getClass(" + className + ")");
            return null;
        }
        debugPrintln("loaded " + className + " from " + which(clazz));
        try {
            if (!SchemaFactory.class.isAssignableFrom(clazz)) {
                throw new ClassCastException(clazz.getName() + " cannot be cast to " + SchemaFactory.class);
            }
            schemaFactory = (SchemaFactory)clazz.newInstance();
        }
        catch (ClassCastException classCastException) {
            debugPrintln("could not instantiate " + clazz.getName());
            if (SchemaFactoryFinder.debug) {
                classCastException.printStackTrace();
            }
            return null;
        }
        catch (IllegalAccessException illegalAccessException) {
            debugPrintln("could not instantiate " + clazz.getName());
            if (SchemaFactoryFinder.debug) {
                illegalAccessException.printStackTrace();
            }
            return null;
        }
        catch (InstantiationException instantiationException) {
            debugPrintln("could not instantiate " + clazz.getName());
            if (SchemaFactoryFinder.debug) {
                instantiationException.printStackTrace();
            }
            return null;
        }
        return schemaFactory;
    }
    
    private boolean isSchemaLanguageSupportedBy(final SchemaFactory factory, final String schemaLanguage, AccessControlContext acc) {
        return (boolean)AccessController.doPrivileged(new PrivilegedAction<Boolean>() {
            @Override
            public Boolean run() {
                return Boolean.valueOf(factory.isSchemaLanguageSupported(schemaLanguage));
            }
        }, acc);
    }
    
    private SchemaFactory findServiceProvider(final String schemaLanguage) {
        assert schemaLanguage != null;
        final AccessControlContext acc = AccessController.getContext();
        try {
            return (SchemaFactory)AccessController.doPrivileged(new PrivilegedAction<SchemaFactory>() {
                @Override
                public SchemaFactory run() {
                    ServiceLoader<SchemaFactory> loader = ServiceLoader.load(SchemaFactoryFinder.SERVICE_CLASS);
                    for (SchemaFactory factory : loader) {
                        if (SchemaFactoryFinder.access$100(SchemaFactoryFinder.this, factory, schemaLanguage, acc)) {
                            return factory;
                        }
                        continue;
                    }
                    return null;
                }
            });
        }
        catch (ServiceConfigurationError error) {
            throw new SchemaFactoryConfigurationError("Provider for " + SchemaFactoryFinder.SERVICE_CLASS + " cannot be created", error);
        }
    }
    
    private static String which(Class<?> clazz) {
        return which(clazz.getName(), clazz.getClassLoader());
    }
    
    private static String which(String classname, ClassLoader loader) {
        String classnameAsResource = classname.replace('.', '/') + ".class";
        if (loader == null) {
            loader = ClassLoader.getSystemClassLoader();
        }
        URL it = SchemaFactoryFinder.ss.getResourceAsURL(loader, classnameAsResource);
        if (it != null) {
            return it.toString();
        }
        return null;
    }
    
    static boolean access$100(SchemaFactoryFinder x0, SchemaFactory x1, String x2, AccessControlContext x3) {
        return x0.isSchemaLanguageSupportedBy(x1, x2, x3);
    }
    
    static {
        SchemaFactoryFinder.debug = false;
        ss = new SecuritySupport();
        cacheProps = new Properties();
        SchemaFactoryFinder.firstTime = true;
        try {
            SchemaFactoryFinder.debug = (SchemaFactoryFinder.ss.getSystemProperty("jaxp.debug") != null);
        }
        catch (Exception unused) {
            SchemaFactoryFinder.debug = false;
        }
        SERVICE_CLASS = SchemaFactory.class;
    }
}
